Estimado/a responsable de la subida de datos de secuenciación del {{ submitting_institution }},

Le informamos que tras el proceso de validación de los datos procesados a fecha {{ batch }}, el número de muestras que no ha sido posible validar es de {{ invalid_count }}.
Las muestras no han superado el proceso de validación debido a errores o faltas de información en los datos proporcionados. Le solicitamos que revise los detalles de cada muestra para subsanar los problemas.
Adjunto a este correo encontrará un documento en formato Excel con información detallada sobre los errores o la información faltante para cada muestra. Le pedimos que proceda a realizar las correcciones necesarias y a reenviar los datos.
En la carpete del SFTP existe una carpeta llamada {{ batch }}_invalid_samples con un excel de metadatos a subsanar con los errores mencionados en el excel adjunto al correo y los .fastq asociados. Una vez subsanados los errores, se puede renombrar la carpeta quitando `_invalid_samples` del nombre, o notificarnoslo en respuesta a este correo para que nos encarguemos nosotros.

{% if additional_info %}
Notas adicionales:
{{ additional_info }}
{% endif %}

En caso de cualquier duda o aclaración, no dude en ponerse en contacto con el equipo técnico empleando este mismo hilo de correos.

Atentamente,

Equipo RELECOV
