Estimado/a:

Le informamos de que ya están disponibles los resultados del análisis correspondiente al batch {{ batch }} en el SFTP.

Puede descargarlos desde la carpeta /{{ submitting_institution_code }}/ANALYSIS_RESULTS, por medio del fichero {{ zip_filename }}.

Para descomprimir el fichero comprimido, le será requerida una contraseña, la cual es {{ password }}.

En caso de cualquier duda o consulta, no dude en ponerse en contacto con el equipo técnico.

Atentamente,

Equipo Relecov